package dev.hephaestus.proximity;

import dev.hephaestus.proximity.api.DataSet;
import dev.hephaestus.proximity.api.Values;
import dev.hephaestus.proximity.api.tasks.DataFinalization;
import dev.hephaestus.proximity.api.tasks.DataPreparation;
import dev.hephaestus.proximity.cards.CardPrototype;
import dev.hephaestus.proximity.api.json.JsonElement;
import dev.hephaestus.proximity.api.json.JsonObject;
import dev.hephaestus.proximity.plugins.Plugin;
import dev.hephaestus.proximity.plugins.PluginHandler;
import dev.hephaestus.proximity.plugins.TaskHandler;
import dev.hephaestus.proximity.plugins.util.Artifact;
import dev.hephaestus.proximity.util.ScriptingUtil;
import dev.hephaestus.proximity.templates.LayerRegistry;
import dev.hephaestus.proximity.templates.RemoteFileSource;
import dev.hephaestus.proximity.templates.TemplateSource;
import dev.hephaestus.proximity.util.Logging;
import dev.hephaestus.proximity.util.RemoteFileCache;
import dev.hephaestus.proximity.util.Result;
import dev.hephaestus.proximity.util.StatefulGraphics;
import dev.hephaestus.proximity.xml.LayerRenderer;
import dev.hephaestus.proximity.xml.RenderableData;
import dev.hephaestus.proximity.xml.XMLUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.quiltmc.json5.JsonReader;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.imageio.*;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;
import javax.imageio.stream.ImageOutputStream;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.net.URLEncoder;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

public final class Proximity {
    public static Logger LOG = LogManager.getLogger("Proximity");

    private final JsonObject options;
    private final TaskHandler taskHandler;
    private final PluginHandler pluginHandler;
    private final LayerRegistry layers;
    private final RemoteFileCache cache;
    private final HashMap<String, Result<JsonObject>> cardInfo = new HashMap<>();

    public Proximity(JsonObject options, TaskHandler taskHandler, PluginHandler pluginHandler, LayerRegistry layers) {
        this.options = options;
        this.taskHandler = taskHandler;
        this.pluginHandler = pluginHandler;
        this.layers = layers;

        try {
            this.cache = RemoteFileCache.load();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public RemoteFileCache getRemoteFileCache() {
        return this.cache;
    }

    public void run(Deque<CardPrototype> prototypes) {
        long startTime = System.currentTimeMillis();

        Result<?> result = this.getCardInfo(prototypes)
                .then(this::renderAndSave)
                .ifError(LOG::error);

        if (!result.isError()) {
            LOG.info("Done! Took {}ms", System.currentTimeMillis() - startTime);
        }
    }

    private Result<Deque<RenderableData>> getCardInfo(Deque<CardPrototype> prototypes) {
        LOG.info("Fetching info for {} cards...", prototypes.size());

        long lastRequest = 0;

       Deque<RenderableData> cards = new ArrayDeque<>();

       int i = 1;
       int prototypeCountLength = Integer.toString(prototypes.size()).length();
       long totalTime = System.currentTimeMillis();

       for (CardPrototype prototype : prototypes) {
           long time = System.currentTimeMillis();

           // We respect Scryfall's wishes and wait between 50-100 seconds between requests.
           if (time - lastRequest < 50) {
               try {
                   LOG.debug("Sleeping for {}ms", time - lastRequest);
                   Thread.sleep(time - lastRequest);
               } catch (InterruptedException e) {
                   e.printStackTrace();
               }
           }

           lastRequest = System.currentTimeMillis();

           getCardInfo(prototype)
                   .ifPresent(raw -> {
                       raw.getAsJsonObject("proximity", "options")
                               .copyAll(this.options)
                               .copyAll(prototype.options());

                       Values.LIST_NAME.set(raw, prototype.listName());

                       for (int j = 0; j < prototype.options().getAsInt("count"); ++j) {
                           int finalJ = j + prototype.number();
                           Values.ITEM_NUMBER.set(raw, finalJ);

                           if (prototype.source().exists("template.xml")) {
                               XMLUtil.load(prototype.source(), "template.xml").ifError(LOG::warn)
                                       .then(root -> this.loadPluginsAndTasks(prototype.source(), root, false))
                                       .then((List<Plugin> plugins) -> {
                                           plugins.forEach(plugin -> plugin.initialize(raw));
                                           return this.runScripts(raw, prototype.overrides());
                                       })
                                       .then((List<JsonObject> list) -> {
                                           list.forEach(card -> XMLUtil.load(prototype.source(), "template.xml").ifError(LOG::warn)
                                                   .then(e -> this.resolveResources(e, prototype.source()))
                                                   .then(e -> this.resolveImports(e, prototype.source()))
                                                   .then(root -> Result.of(new RenderableData(this, prototype.source(), root, card)))
                                                   .then(renderable -> {
                                                       cards.add(renderable);
                                                       return Result.of((Void) null);
                                                   })
                                           );

                                           return Result.of((Void) null);
                                       });
                           } else {
                               LOG.warn("template.xml not found for template {}", prototype.source().getTemplateName());
                           }
                       }
                   })
                   .ifError(LOG::warn);
           System.out.printf("%" + prototypeCountLength + "d/%d\r", i++, prototypes.size());
       }

        System.out.printf("%" + prototypeCountLength + "d/%d%n", i - 1, prototypes.size());

        LOG.info("Successfully found {} cards. Took {}ms", cards.size(), System.currentTimeMillis() - totalTime);

        return Result.of(cards);
    }

    private Result<Element> resolveResources(Element root, TemplateSource.Compound source) {
        NodeList resourceList = root.getElementsByTagName("Resources");

        for (int i = 0; i < resourceList.getLength(); ++i) {
            Node r = resourceList.item(i);

            if (r instanceof Element) {
                NodeList resources = r.getChildNodes();

                for (int j = 0; j < resources.getLength(); ++j) {
                    r = resources.item(j);

                    if (r instanceof Element resource) {
                        String location = resource.getAttribute("location") + "/" + resource.getAttribute("version");

                        //noinspection SwitchStatementWithTooFewBranches May add more kinds of resources later
                        switch (resource.getAttribute("type")) {
                            case "assets" -> {
                                if (cache != null) {
                                    source.wrapped.add(new RemoteFileSource(cache, location));
                                }
                            }
                        }
                    }
                }
            }
        }

        return Result.of(root);
    }

    private Result<Element> resolveImports(Element root, TemplateSource source) {
        Node firstChild = root.getFirstChild();
        Document document = root.getOwnerDocument();
        NodeList importsList = root.getElementsByTagName("Imports");

        for (int i = 0; i < importsList.getLength(); ++i) {
            Node n = importsList.item(i);

            if (n instanceof Element) {
                NodeList imports = n.getChildNodes();

                for (int j = 0; j < imports.getLength(); ++j) {
                    n = imports.item(j);

                    if (n instanceof Element element) {
                        String src = element.getAttribute("src") + ".xml";

                        XMLUtil.load(source, src)
                                .ifError(LOG::warn)
                                .then(imported -> {
                                    NodeList children = imported.getChildNodes();

                                    for (int k = 0; k < children.getLength(); ++k) {
                                        Node m = children.item(k);

                                        if (m instanceof Element e) {
                                            root.insertBefore(document.adoptNode(e), firstChild);
                                        }
                                    }

                                    return Result.of((Void) null);
                                });
                    }
                }
            }
        }

        return Result.of(root);
    }

    private Result<List<Plugin>> loadPluginsAndTasks(TemplateSource source, Element root, boolean help) {
        NodeList pluginBlocks = root.getElementsByTagName("Plugins");
        List<Plugin> plugins = new ArrayList<>();

        for (int i = 0; i < pluginBlocks.getLength(); ++i) {
            Node n = pluginBlocks.item(i);

            if (n instanceof Element e) {
                NodeList imports = e.getElementsByTagName("Import");

                for (int j = 0; j < imports.getLength(); ++j) {
                    n = imports.item(j);

                    if (n instanceof Element importElement) {
                        String repository = importElement.getAttribute("repository");
                        String group = importElement.getAttribute("group");
                        String artifact = importElement.getAttribute("artifact");
                        String versionRange = importElement.getAttribute("version");

                        Result<Plugin> result = this.pluginHandler.loadPlugin(Artifact.create(
                                repository, group, artifact, versionRange
                        ), this.taskHandler, help);

                        if (result.isError()) {
                            return result.unwrap();
                        } else {
                            plugins.add(result.get());
                        }
                    }
                }
            }
        }

        NodeList tasksTags = root.getElementsByTagName("Tasks");

        for (int i = 0; i < tasksTags.getLength(); ++i) {
            Node n = tasksTags.item(i);

            if (n instanceof Element tasks) {
                NodeList taskTags = tasks.getChildNodes();

                for (int j = 0; j < taskTags.getLength(); ++j) {
                    n = taskTags.item(j);

                    if (n instanceof Element task && this.taskHandler.contains(task.getTagName())) {
                        Function<Object[], Object> function = ScriptingUtil.getFunction(source, task.getAttribute("location"));
                        String taskName = task.getTagName();

                        if (task.hasAttribute("name")) {
                            this.taskHandler.put(taskName, task.getAttribute("name"), function);
                        } else {
                            this.taskHandler.put(taskName, function);
                        }
                    }
                }
            }
        }

        return Result.of(plugins);
    }

    public Result<List<JsonObject>> runScripts(JsonObject raw, JsonObject overrides) {
        List<JsonObject> list = new ArrayList<>();

        list.add(raw);

        DataSet data = new DataSet(list);
        List<DataPreparation> preparations = this.taskHandler.getTasks(DataPreparation.DEFINITION);

        for (var task : preparations) {
            task.apply(this.taskHandler, data, overrides);
        }

        List<DataFinalization> finalizations = this.taskHandler.getTasks(DataFinalization.DEFINITION);

        for (var task : finalizations) {
            task.apply();
        }

        return Result.of(list);
    }

    private Result<JsonObject> getCardInfo(CardPrototype prototype) {
        StringBuilder uri = new StringBuilder("https://api.scryfall.com/cards/named?");

        uri.append("fuzzy=").append(URLEncoder.encode(prototype.cardName(), StandardCharsets.UTF_8));

        if (prototype.options().has("set_code")) {
            if (prototype.options().has("collector_number")) {
                uri = new StringBuilder("https://api.scryfall.com/cards/")
                        .append(prototype.options().getAsString("set_code").toLowerCase(Locale.ROOT))
                        .append("/")
                        .append(prototype.options().getAsString("collector_number"))
                ;
            } else {
                uri.append("&set=").append(prototype.options().getAsString("set_code"));
            }
        }

        String string = uri.toString();

        return this.cardInfo.computeIfAbsent(string, s -> {
            HttpClient client = HttpClient.newHttpClient();
            HttpRequest request = HttpRequest.newBuilder()
                    .uri(URI.create(s))
                    .GET()
                    .build();

            try {
                HttpResponse<String> response = client.send(request, HttpResponse.BodyHandlers.ofString());

                if (response.statusCode() == 200) {
                    String responseBody = response.body();
                    JsonObject cardInfo = JsonObject.parseObject(JsonReader.json5(responseBody));

                    return Result.of(cardInfo);
                } else {
                    StringBuilder message = new StringBuilder("Could not find card ").append(prototype.cardName());

                    if (prototype.options().has("set_code")) {
                        message.append(" (")
                                .append(prototype.options().getAsString("set_code").toUpperCase(Locale.ROOT))
                                .append(")");

                        if (prototype.options().has("collector_number")) {
                            message.append(" #")
                                    .append(prototype.options().getAsString("collector_number").toUpperCase(Locale.ROOT));
                        }
                    }

                    JsonObject body = JsonObject.parseObject(JsonReader.json(response.body()));
                    String details = "[" + response.statusCode() + "] " + (body.has("details") ? body.getAsString("details") : "");

                    return Result.error("%s: %s", message.toString(), details);
                }
            } catch (Exception e) {
                return Result.error(e.getMessage());
            }
        });
    }

    private Result<AtomicInteger> renderAndSave(Deque<RenderableData> cards) {
        int countStrLen = Integer.toString(cards.size()).length();
        int threadCount = Integer.min(options.has("threads") ? Integer.parseInt(options.getAsString("threads")) : 10, cards.size());

        if (threadCount > 0) {
            AtomicInteger finishedCards = new AtomicInteger();
            ExecutorService executor = Executors.newFixedThreadPool(threadCount);
            Deque<String> errors = new ConcurrentLinkedDeque<>();

            LOG.info("Rendering {} cards on {} threads", cards.size(), threadCount);

            for (RenderableData card : cards) {
                if (threadCount > 1) {
                    //                    this.render(card, finishedCards, errors, countStrLen, finalI, cards.size());
                    executor.submit(() -> this.render(card, finishedCards, errors, countStrLen, cards.size()));
                } else {
                    this.render(card, finishedCards, errors, countStrLen, cards.size());
                }

            }

            try {
                executor.shutdown();

                //noinspection StatementWithEmptyBody
                while (!executor.awaitTermination(Integer.MAX_VALUE, TimeUnit.NANOSECONDS)) {

                }

                return errors.isEmpty()
                        ? Result.of(finishedCards)
                        : Result.error("Error rendering cards:\n\t%s", String.join("\n\t", errors));
            } catch (InterruptedException e) {
                return Result.error(e.getMessage());
            }
        }

        return Result.error("No cards to render");
    }

    private void render(RenderableData card, AtomicInteger finishedCards, Deque<String> errors, int countStrLen, int cardCount) {
        long cardTime = System.currentTimeMillis();

        String name = card.getName();

        try {
            BufferedImage image = new BufferedImage(card.getWidth(), card.getHeight(), BufferedImage.TYPE_INT_ARGB);

            Result<Void> result = card.render(new StatefulGraphics(image)).ifError(errors::add);

            if (result.isOk()) {
                Path path = Path.of("images");

                for (JsonElement element : card.getAsJsonArray("proximity", "path")) {
                    path = path.resolve(element.getAsString());
                }

                path = path.resolveSibling(path.getFileName().toString() + ".png");

                this.save(image, path);

                finishedCards.incrementAndGet();
                LOG.info(String.format("%" + countStrLen + "d/%" + countStrLen + "d  %5dms  %-55s {}SAVED{}", finishedCards.get(), cardCount, System.currentTimeMillis() - cardTime, name), Logging.ANSI_GREEN, Logging.ANSI_RESET);
            } else {
                LOG.error(String.format("%" + countStrLen + "d/%" + countStrLen + "d  %5dms  %-55s {}FAILED{}", finishedCards.get(), cardCount, System.currentTimeMillis() - cardTime, name), Logging.ANSI_RED, Logging.ANSI_RESET);
            }
        } catch (Throwable throwable) {
            LOG.error(String.format("%" + countStrLen + "d/%" + countStrLen + "d  %5dms  %-55s {}FAILED{}", finishedCards.get(), cardCount, System.currentTimeMillis() - cardTime, name), Logging.ANSI_RED, Logging.ANSI_RESET);
            LOG.error(throwable.getMessage());

            for (StackTraceElement element : throwable.getStackTrace()) {
                LOG.debug(element);
            }
        }
    }

    private void save(BufferedImage image, Path path) throws Throwable {
        if (!Files.isDirectory(path.getParent())) {
            Files.createDirectories(path.getParent());
        }

        ImageWriter writer = ImageIO.getImageWritersByFormatName("png").next();

        ImageWriteParam param = writer.getDefaultWriteParam();
        ImageTypeSpecifier typeSpecifier = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_INT_ARGB);
        IIOMetadata metadata = writer.getDefaultImageMetadata(typeSpecifier, param);

        IIOMetadataNode discriminate = new IIOMetadataNode("tEXtEntry");
        Random random = new Random();
        byte[] bytes = new byte[128];

        discriminate.setAttribute("keyword", "ProximityDiscriminate");
        random.nextBytes(bytes);
        discriminate.setAttribute("value", new String(Base64.getEncoder().encode(bytes)));

        IIOMetadataNode text = new IIOMetadataNode("tEXt");
        text.appendChild(discriminate);

        IIOMetadataNode root = new IIOMetadataNode("javax_imageio_png_1.0");
        root.appendChild(text);

        metadata.mergeTree("javax_imageio_png_1.0", root);

        OutputStream stream = Files.newOutputStream(path);
        ImageOutputStream imageOutputStream = ImageIO.createImageOutputStream(stream);

        writer.setOutput(imageOutputStream);
        writer.write(metadata, new IIOImage(image, null, metadata), param);
        imageOutputStream.close();
        stream.close();
    }

    public Map<String, LayerRenderer> createLayerRenderers(RenderableData data) {
        return this.layers.create(data);
    }

    public TaskHandler getTaskHandler() {
        return this.taskHandler;
    }

    public void help(TemplateSource.Compound source) {
        JsonObject json = new JsonObject();

        Values.HELP.set(json, true);

        if (source.exists("template.xml")) {
            XMLUtil.load(source, "template.xml").ifError(LOG::warn)
                    .then(root -> {
                        this.loadPluginsAndTasks(source, root, true);
                        new RenderableData(this, source, root, json).parseOptions();
                        return null;
                    });
        } else {
            LOG.warn("template.xml not found for template {}", source.getTemplateName());
        }
    }
}
